from __future__ import annotations

from .genome import PackedGenomicData, SequencePacker64
from .outbreak import build_pairwise_case_table, simulate_epidemic_dates, simulate_genomic_sequences

__all__ = [
    "build_pairwise_case_table",
    "simulate_epidemic_dates",
    "simulate_genomic_sequences",
    "PackedGenomicData",
    "SequencePacker64",
]



from __future__ import annotations

import math
import sys
import unittest
from pathlib import Path
from unittest.mock import patch

import numpy as np

ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
if str(SRC) not in sys.path:
    sys.path.insert(0, str(SRC))

from epilink import (  # noqa: E402
    BaseTransmissionProfile,
    InfectiousnessToTransmission,
    NaturalHistoryParameters,
    SymptomOnsetToTransmission,
)


class ConstantDensityProfile(BaseTransmissionProfile):
    def __init__(self, **kwargs) -> None:
        kwargs.setdefault("grid_min_days", 0.0)
        kwargs.setdefault("grid_max_days", 2.0)
        kwargs.setdefault("grid_points", 5)
        super().__init__(**kwargs)
        self.pdf_calls = 0

    def pdf(self, times_in_days: np.typing.ArrayLike) -> np.ndarray:
        self.pdf_calls += 1
        evaluation_points = np.asarray(times_in_days, dtype=float)
        density = np.where(
            (evaluation_points >= self.grid_min_days) & (evaluation_points <= self.grid_max_days),
            0.5,
            0.0,
        )
        return density[()] if evaluation_points.ndim == 0 else density

    def rvs(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        sample_shape = (size,) if isinstance(size, int) else size
        return np.full(sample_shape, 1.0, dtype=float)


class ZeroDensityProfile(BaseTransmissionProfile):
    def __init__(self, **kwargs) -> None:
        super().__init__(grid_min_days=0.0, grid_max_days=2.0, grid_points=5, **kwargs)

    def pdf(self, times_in_days: np.typing.ArrayLike) -> np.ndarray:
        evaluation_points = np.asarray(times_in_days, dtype=float)
        density = np.zeros_like(evaluation_points)
        return density[()] if evaluation_points.ndim == 0 else density

    def rvs(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        sample_shape = (size,) if isinstance(size, int) else size
        return np.zeros(sample_shape, dtype=float)


class TestBaseTransmissionProfile(unittest.TestCase):
    def test_invalid_grid_configuration_raises_value_error(self) -> None:
        with self.assertRaises(ValueError):
            ConstantDensityProfile(grid_min_days=2.0, grid_max_days=1.0)

        with self.assertRaises(ValueError):
            ConstantDensityProfile(grid_points=1)

    def test_cdf_uses_cached_numerical_grid(self) -> None:
        profile = ConstantDensityProfile()

        first = profile.cdf(np.array([0.0, 1.0, 2.0]))
        second = profile.cdf(np.array([0.5, 1.5]))

        np.testing.assert_allclose(first, np.array([0.0, 0.5, 1.0]))
        np.testing.assert_allclose(second, np.array([0.25, 0.75]))
        self.assertEqual(profile.pdf_calls, 1)

    def test_cdf_falls_back_to_linear_interpolation_when_density_has_no_mass(self) -> None:
        profile = ZeroDensityProfile()

        cdf = profile.cdf(np.array([0.0, 1.0, 2.0]))

        np.testing.assert_allclose(cdf, np.array([0.0, 0.5, 1.0]))

    def test_sample_incubation_periods_adds_latent_and_presymptomatic_periods(self) -> None:
        profile = ConstantDensityProfile()
        latent_periods = np.array([1.0, 2.0, 3.0])
        presymptomatic_periods = np.array([0.25, 0.5, 0.75])

        with (
            patch.object(profile, "sample_latent_periods", return_value=latent_periods),
            patch.object(
                profile,
                "sample_presymptomatic_periods",
                return_value=presymptomatic_periods,
            ),
        ):
            incubation_periods = profile.sample_incubation_periods(size=3)

        np.testing.assert_allclose(incubation_periods, latent_periods + presymptomatic_periods)

    def test_mean_and_repr_reflect_profile_configuration(self) -> None:
        profile = ConstantDensityProfile()

        self.assertTrue(math.isclose(profile.mean(), 1.0, rel_tol=1e-3))
        self.assertIn("ConstantDensityProfile", repr(profile))
        self.assertIn("grid_min_days=0.0", repr(profile))
        self.assertIn("grid_max_days=2.0", repr(profile))

    def test_zero_mass_profile_has_nan_mean(self) -> None:
        profile = ZeroDensityProfile()

        self.assertTrue(math.isnan(profile.mean()))

    def test_strict_clock_sampling_and_expected_mutations_are_deterministic(self) -> None:
        parameters = NaturalHistoryParameters(
            substitution_rate=2.0e-3,
            relaxation=0.0,
            genome_length=1000,
        )
        profile = ConstantDensityProfile(parameters=parameters, rng_seed=123)

        expected_rate_per_day = (parameters.substitution_rate * parameters.genome_length) / 365.0
        clock_rate = profile.sample_clock_rate(size=(2, 2))
        expected_mutations = profile.expected_mutations(np.array([0.0, 5.0, 10.0]))

        np.testing.assert_allclose(clock_rate, np.full((2, 2), expected_rate_per_day))
        np.testing.assert_allclose(
            expected_mutations,
            np.array([0.0, 5.0 * expected_rate_per_day, 10.0 * expected_rate_per_day]),
        )


class TestInfectiousnessToTransmission(unittest.TestCase):
    def test_pdf_returns_all_zeros_when_every_time_is_negative(self) -> None:
        profile = InfectiousnessToTransmission(
            rng_seed=123,
            integration_grid_points=64,
            sampling_grid_points=64,
        )

        values = profile.pdf(np.array([-3.0, -1.5, -0.2]))

        np.testing.assert_allclose(values, np.zeros(3))

    def test_pdf_returns_zero_for_negative_times_and_scalar_for_scalar_input(self) -> None:
        profile = InfectiousnessToTransmission(
            rng_seed=123,
            integration_grid_points=64,
            sampling_grid_points=64,
        )

        values = profile.pdf(np.array([-2.0, -0.5, 0.0, 1.0]))
        scalar = profile.pdf(0.0)

        np.testing.assert_allclose(values[:2], np.array([0.0, 0.0]))
        self.assertTrue(np.all(values >= 0.0))
        self.assertIsInstance(scalar, float)

    def test_rvs_respects_shape_and_grid_bounds(self) -> None:
        profile = InfectiousnessToTransmission(
            rng_seed=123,
            integration_grid_points=64,
            sampling_grid_points=64,
        )

        draws = profile.rvs(size=(3, 4))

        self.assertEqual(draws.shape, (3, 4))
        self.assertTrue(np.all(draws >= profile.grid_min_days))
        self.assertTrue(np.all(draws <= profile.grid_max_days))

    def test_sample_generation_intervals_adds_latent_periods_and_transmission_times(self) -> None:
        profile = InfectiousnessToTransmission(rng_seed=123)
        latent_periods = np.array([1.0, 2.0, 3.0])
        transmission_times = np.array([0.5, 0.75, 1.25])

        with (
            patch.object(profile, "sample_latent_periods", return_value=latent_periods),
            patch.object(profile, "rvs", return_value=transmission_times),
        ):
            generation_intervals = profile.sample_generation_intervals(size=3)

        np.testing.assert_allclose(generation_intervals, latent_periods + transmission_times)


class TestSymptomOnsetToTransmission(unittest.TestCase):
    def test_pdf_matches_documented_piecewise_definition(self) -> None:
        profile = SymptomOnsetToTransmission(rng_seed=123, grid_points=128)
        points = np.array([-2.0, -0.5, 0.0, 1.5])
        parameters = profile.parameters

        expected = np.where(
            points < 0.0,
            parameters.transmission_rate_ratio
            * parameters.infectiousness_normalisation
            * (1.0 - profile.presymptomatic.cdf(-points)),
            parameters.infectiousness_normalisation * (1.0 - profile.symptomatic.cdf(points)),
        )

        np.testing.assert_allclose(profile.pdf(points), expected)
        self.assertIsInstance(profile.pdf(0.0), float)

    def test_cdf_and_rvs_respect_configured_bounds(self) -> None:
        profile = SymptomOnsetToTransmission(rng_seed=123, grid_points=128)

        draws = profile.rvs(size=20)

        self.assertEqual(profile.cdf(profile.grid_min_days - 1.0), 0.0)
        self.assertEqual(profile.cdf(profile.grid_max_days + 1.0), 1.0)
        self.assertEqual(draws.shape, (20,))
        self.assertTrue(np.all(draws >= profile.grid_min_days))
        self.assertTrue(np.all(draws <= profile.grid_max_days))



from __future__ import annotations

from dataclasses import asdict, dataclass
from numbers import Real


@dataclass(frozen=True, slots=True, kw_only=True)
class NaturalHistoryParameters:
    r"""
    Parameters for the E/P/I variable infectiousness model [1].

    The model assumes stage durations follow Gamma distributions using the
    shape-scale parameterization.

    Attributes
    ----------
    incubation_shape : float, default=5.807
        Shape parameter :math:`k_{inc}` of the incubation period distribution.
        Must be a positive real number (dimensionless).
    incubation_scale : float, default=0.948
        Scale parameter :math:`\theta_{inc}` of the incubation period distribution.
        Must be a positive real number (days).
    latent_shape : float, default=3.38
        Shape parameter :math:`k_E` of the latent (``E``) stage.
        Must satisfy :math:`0 < k_E < k_{inc}` (dimensionless).
    symptomatic_rate : float, default=0.37
        Symptomatic removal rate :math:`\mu`.
        For :math:`k_I = 1`, the mean symptomatic duration is :math:`1/\mu`.
        Must be a positive real number (1/day).
    symptomatic_shape : float, default=1.0
        Shape parameter :math:`k_I` of the symptomatic infectious (``I``) stage.
        Must be a positive real number (dimensionless).
    transmission_rate_ratio : float, default=2.29
        Ratio of transmission rates in presymptomatic (``P``) and symptomatic
        infectious (``I``) stages, :math:`\alpha` for ``P`` compared with ``I``.
        Must be a positive real number (dimensionless).
    testing_delay_shape : float, default=2.0
        Shape parameter for the Gamma-distributed testing delay.
        Must be a positive real number (dimensionless).
    testing_delay_scale : float, default=1.0
        Scale parameter for the Gamma-distributed testing delay.
        Must be a positive real number (days).
    substitution_rate : float, default=1e-3
        Median substitution rate in substitutions per site per year.
        Must be a positive real number (substitutions/site/year).
    relaxation: float, default=0.33
        Lognormal standard deviation for relaxed clock branch-specific rates.
        Must be a non-negative real number (dimensionless).
        Set to 0 to use strict clock.
    genome_length: int, default=29903
        Genome length in sites. Must be a positive integer.

    Notes
    -----
    Derived parameters are exposed as read-only properties.

    `presymptomatic_shape`
        :math:`k_P = k_{inc} - k_E` for the presymptomatic infectious (``P``) stage.

    `symptomatic_scale`
        :math:`\theta_I = 1 / (k_I \cdot \mu)` for symptomatic stage duration (days).

    `incubation_rate`
        :math:`\lambda = 1 / (k_{inc} \cdot \theta_{inc})` (1/day).

    `infectiousness_normalisation`
        Normalisation constant :math:`C` such that the infectiousness profile integrates
        to unity over the infectious period:

        .. math::

            C = \frac{k_{inc} \cdot \lambda \cdot \mu}
            {\alpha \cdot k_P \cdot \mu + k_{inc} \cdot \lambda}

    `presymptomatic_transmission_fraction`
        Presymptomatic transmission fraction :math:`q_P \in [0, 1]`:

        .. math::

            q_P = \frac{\alpha \cdot k_P \cdot \mu}
            {\alpha \cdot k_P \cdot \mu + k_{inc} \cdot \lambda}

    References
    ----------
    .. [1] Hart, William S., Philip K. Maini, and Robin N. Thompson (2021)
        High infectiousness immediately before COVID-19 symptom onset
        highlights the importance of continued contact tracing. eLife 10:e65534.
    """

    incubation_shape: float = 5.807
    incubation_scale: float = 0.948
    latent_shape: float = 3.38
    symptomatic_rate: float = 0.37
    symptomatic_shape: float = 1.0
    transmission_rate_ratio: float = 2.29
    testing_delay_shape: float = 2.0
    testing_delay_scale: float = 1.0
    substitution_rate: float = 1e-3
    relaxation: float = 0.33
    genome_length: int = 29903

    def __post_init__(self) -> None:
        positive_real_fields = (
            "incubation_shape",
            "incubation_scale",
            "latent_shape",
            "symptomatic_rate",
            "symptomatic_shape",
            "transmission_rate_ratio",
            "testing_delay_shape",
            "testing_delay_scale",
            "substitution_rate",
            "genome_length",
        )
        for field_name in positive_real_fields:
            value = getattr(self, field_name)
            if isinstance(value, bool) or not isinstance(value, Real):
                raise TypeError(f"{field_name} must be a real number.")
            if value <= 0:
                raise ValueError(f"{field_name} must be positive.")

        if isinstance(self.relaxation, bool) or not isinstance(self.relaxation, Real):
            raise TypeError("relaxation must be a real number.")
        if self.relaxation < 0:
            raise ValueError("relaxation must be non-negative.")

        if self.latent_shape >= self.incubation_shape:
            raise ValueError(
                "latent_shape must be < incubation_shape (so presymptomatic_shape is positive)."
            )

    @property
    def presymptomatic_shape(self) -> float:
        return self.incubation_shape - self.latent_shape

    @property
    def symptomatic_scale(self) -> float:
        return 1.0 / (self.symptomatic_shape * self.symptomatic_rate)

    @property
    def incubation_rate(self) -> float:
        return 1.0 / (self.incubation_shape * self.incubation_scale)

    @property
    def infectiousness_normalisation(self) -> float:
        numerator = self.incubation_shape * self.incubation_rate * self.symptomatic_rate
        denominator = (
            self.transmission_rate_ratio * self.presymptomatic_shape * self.symptomatic_rate
            + self.incubation_shape * self.incubation_rate
        )
        return numerator / denominator

    @property
    def presymptomatic_transmission_fraction(self) -> float:
        num = self.transmission_rate_ratio * self.presymptomatic_shape * self.symptomatic_rate
        den = num + (self.incubation_shape * self.incubation_rate)
        return num / den

    def to_dict(self) -> dict[str, float]:
        """Return constructor parameters as a plain dictionary."""
        return asdict(self)


__all__ = ["NaturalHistoryParameters"]



from __future__ import annotations

import textwrap
from typing import TypeAlias

import numpy as np
import numpy.typing as npt

NDArrayInt8: TypeAlias = npt.NDArray[np.int8]
NDArrayUInt64: TypeAlias = npt.NDArray[np.uint64]
NDArrayInt32: TypeAlias = npt.NDArray[np.int32]

_PACK_SHIFTS = np.arange(62, -2, -2, dtype=np.uint64)
_UINT64_BYTES = np.dtype(np.uint64).itemsize
_DEFAULT_HAMMING_WORKING_SET_BYTES = 32 * 1024 * 1024
_XOR_BYTE_TO_HAMMING = np.array(
    [bin((byte | (byte >> 1)) & 0x55).count("1") for byte in range(256)],
    dtype=np.uint8,
)


class SequencePacker64:
    """
    Utility engine for 64-bit packing and Hamming distance calculation.
    """

    @staticmethod
    def _resolve_block_size(
        num_sequences: int,
        bytes_per_sequence: int,
        block_size: int | None,
    ) -> int:
        if block_size is not None:
            return max(1, min(num_sequences, int(block_size)))
        if num_sequences <= 1 or bytes_per_sequence == 0:
            return max(1, num_sequences)

        max_pairs = max(
            1,
            _DEFAULT_HAMMING_WORKING_SET_BYTES // max(1, 2 * bytes_per_sequence),
        )
        return max(1, min(num_sequences, int(np.sqrt(max_pairs))))

    @staticmethod
    def pack_u64(arr: NDArrayInt8) -> NDArrayUInt64:
        """
        Pack 2-bit nucleotide arrays into 64-bit blocks.
        """

        n_sequences, sequence_length = arr.shape
        n_blocks = (sequence_length + 31) // 32
        out = np.zeros((n_sequences, n_blocks), dtype=np.uint64)

        if n_blocks == 0:
            return out

        padded = np.zeros((n_sequences, n_blocks * 32), dtype=np.uint8)
        padded[:, :sequence_length] = arr
        grouped = padded.reshape(n_sequences, n_blocks, 32)

        for position, shift in enumerate(_PACK_SHIFTS):
            out |= grouped[:, :, position].astype(np.uint64) << shift

        return out

    @staticmethod
    def hamming64(
        packed: NDArrayUInt64,
        block_size: int | None = None,
    ) -> NDArrayInt32:
        """
        Compute pairwise Hamming distances for packed sequences.

        Parameters
        ----------
        packed : numpy.ndarray
            Packed 64-bit representation of the sequences.
        block_size : int, optional
            Number of sequences to compare per row/column chunk. Smaller values
            reduce peak memory for large pairwise calculations.
        """

        packed = np.ascontiguousarray(packed, dtype=np.uint64)
        n_sequences, n_blocks = packed.shape
        distances = np.zeros((n_sequences, n_sequences), dtype=np.int32)

        if n_sequences == 0 or n_blocks == 0:
            return distances

        packed_bytes = packed.view(np.uint8).reshape(n_sequences, n_blocks * _UINT64_BYTES)
        step = SequencePacker64._resolve_block_size(
            n_sequences,
            packed_bytes.shape[1],
            block_size,
        )

        for row_start in range(0, n_sequences, step):
            row_stop = min(row_start + step, n_sequences)
            left = packed_bytes[row_start:row_stop]

            for col_start in range(row_start, n_sequences, step):
                col_stop = min(col_start + step, n_sequences)
                right = packed_bytes[col_start:col_stop]
                xor_bytes = np.bitwise_xor(left[:, np.newaxis, :], right[np.newaxis, :, :])
                block = _XOR_BYTE_TO_HAMMING[xor_bytes].sum(axis=2, dtype=np.int32)

                distances[row_start:row_stop, col_start:col_stop] = block
                if col_start != row_start:
                    distances[col_start:col_stop, row_start:row_stop] = block.T

        return distances


class PackedGenomicData:
    """
    Container for packed genomic sequences and lookup metadata.
    """

    def __init__(
        self,
        int8_matrix: NDArrayInt8,
        original_length: int,
        node_map: dict[str, int],
        base_map: dict[int, str],
    ):
        self.n_seqs: int
        self.original_length: int
        self.node_to_idx: dict[str, int]
        self.idx_to_node: dict[int, str]
        self.bases_map: dict[int, str]
        self.packed_u64: NDArrayUInt64

        if not np.issubdtype(int8_matrix.dtype, np.integer):
            raise TypeError("int8_matrix must be an integer array with values in {0, 1, 2, 3}.")
        if int8_matrix.size:
            min_val = int8_matrix.min()
            max_val = int8_matrix.max()
            if min_val < 0 or max_val > 3:
                raise ValueError("int8_matrix contains values outside {0, 1, 2, 3}.")
        if int8_matrix.dtype != np.int8 or not int8_matrix.flags["C_CONTIGUOUS"]:
            int8_matrix = np.ascontiguousarray(int8_matrix, dtype=np.int8)

        self.n_seqs, _ = int8_matrix.shape
        self.original_length = original_length
        self.node_to_idx = node_map
        self.idx_to_node = {v: k for k, v in node_map.items()}
        self.bases_map = base_map
        self.packed_u64 = SequencePacker64.pack_u64(int8_matrix)

    def compute_hamming_distances(
        self,
        block_size: int | None = None,
    ) -> NDArrayInt32:
        """
        Compute the pairwise Hamming distance matrix.

        Parameters
        ----------
        block_size : int, optional
            Number of sequences to compare per row/column chunk. Smaller values
            reduce peak memory for large pairwise calculations.
        """

        packed = (
            self.packed_u64
            if self.packed_u64.flags["C_CONTIGUOUS"]
            else np.ascontiguousarray(self.packed_u64)
        )
        return SequencePacker64.hamming64(packed, block_size=block_size)

    def write_fasta(self, filepath: str) -> None:
        """
        Write unpacked sequences to a FASTA file.
        """

        base_lookup = np.array([self.bases_map[idx] for idx in range(4)], dtype="<U1")

        with open(filepath, "w") as f:
            for i in range(self.n_seqs):
                blocks = self.packed_u64[i]
                L = self.original_length
                unpacked = np.empty((len(blocks), 32), dtype=np.int8)

                for position, shift in enumerate(_PACK_SHIFTS):
                    unpacked[:, position] = ((blocks >> shift) & np.uint64(3)).astype(np.int8)

                seq = unpacked.reshape(len(blocks) * 32)[:L]
                seq_str = "".join(base_lookup[seq])

                header = f">{self.idx_to_node[i]}"
                body = textwrap.fill(seq_str, width=100)
                f.write(f"{header}\n{body}\n")


__all__ = ["PackedGenomicData", "SequencePacker64"]



from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any

import numpy as np
from numpy.random import Generator
from numpy.typing import ArrayLike

from epilink.model.profiles import InfectiousnessToTransmission


@dataclass(frozen=True, slots=True)
class Scenario:
    """Latent transmission scenario."""

    kind: str
    intermediates: int | None = None
    branch_to_i: int | None = None
    branch_to_j: int | None = None

    def label(self) -> str:
        if self.kind == "ad":
            return f"ad({self.intermediates})"
        return f"ca({self.branch_to_i},{self.branch_to_j})"


class PairwiseCompatibilityModel:
    """
    Vectorized pairwise compatibility scorer for a fixed target subset.

    Parameters
    ----------
    draws_by_scenario :
        Cached Monte Carlo draws keyed by scenario label.
    target_labels :
        Canonical scenario labels included in the target subset.
    """

    def __init__(
        self,
        draws_by_scenario: dict[str, dict[str, np.ndarray]],
        target_labels: Iterable[str],
    ) -> None:
        self.target_labels = tuple(target_labels)
        if not self.target_labels:
            raise ValueError("target_labels must contain at least one scenario.")

        self._sorted_time_draws = tuple(
            np.sort(np.asarray(draws_by_scenario[label]["time_draws"], dtype=float))
            for label in self.target_labels
        )
        self._sorted_genetic_draws = tuple(
            np.sort(np.asarray(draws_by_scenario[label]["genetic_draws"], dtype=float))
            for label in self.target_labels
        )

    @staticmethod
    def _broadcast_observations(
        sample_time_difference: ArrayLike,
        genetic_distance: ArrayLike,
    ) -> tuple[np.ndarray, np.ndarray, tuple[int, ...]]:
        sample_time_values, genetic_values = np.broadcast_arrays(
            np.asarray(sample_time_difference, dtype=float),
            np.asarray(genetic_distance, dtype=float),
        )
        return (
            sample_time_values.reshape(-1),
            genetic_values.reshape(-1),
            sample_time_values.shape,
        )

    @staticmethod
    def _percentile_scores(
        observed_values: np.ndarray,
        simulated_sorted_values: np.ndarray,
    ) -> np.ndarray:
        return (
            np.searchsorted(simulated_sorted_values, observed_values, side="right").astype(float)
            / simulated_sorted_values.size
        )

    @staticmethod
    def _compatibility_from_percentiles(percentiles: np.ndarray) -> np.ndarray:
        return 1.0 - 2.0 * np.abs(percentiles - 0.5)

    def score(
        self,
        sample_time_difference: ArrayLike,
        genetic_distance: ArrayLike,
    ) -> np.ndarray:
        """
        Compute target-subset compatibility for arraylike observations.

        Returns
        -------
        numpy.ndarray
            Compatibility scores with the broadcast shape of the inputs.
        """

        sample_time_values, genetic_values, output_shape = self._broadcast_observations(
            sample_time_difference=sample_time_difference,
            genetic_distance=genetic_distance,
        )
        target_scores = np.zeros(sample_time_values.shape, dtype=float)

        for time_draws, genetic_draws in zip(
            self._sorted_time_draws,
            self._sorted_genetic_draws,
            strict=True,
        ):
            time_percentiles = self._percentile_scores(sample_time_values, time_draws)
            genetic_percentiles = self._percentile_scores(genetic_values, genetic_draws)
            target_scores += self._compatibility_from_percentiles(
                time_percentiles
            ) * self._compatibility_from_percentiles(genetic_percentiles)

        return target_scores.reshape(output_shape)

    __call__ = score


class EpiLink:
    """
    Pairwise scenario compatibility scoring.

    Parameters
    ----------
    transmission_profile :
        Instance of ``InfectiousnessToTransmission`` providing generation-interval,
        testing-delay, and per-day substitution-rate samples.
    maximum_depth :
        Maximum hidden-depth parameter.
    mc_samples :
        Number of Monte Carlo draws precomputed for each scenario.
    target :
        Scenario label or subset of scenario labels returned as the main score.
        Accepts canonical labels such as ``ad(0)`` and ``ca(0,0)``, or
        ``Scenario`` objects.
        When multiple targets are provided, the reported target score is the
        sum of the compatibilities across the configured subset.
    mutation_process :
        Either ``"deterministic"`` to compare against expected mutation counts
        or ``"stochastic"`` to compare against Poisson mutation-count draws.
    rng :
        Random number generator used for stochastic mutation-count draws. When
        omitted, the transmission profile generator is reused.

    Notes
    -----
    Use :meth:`pairwise_model` to obtain a dedicated cached scorer for a
    target subset. The resulting model accepts scalar or arraylike inputs and
    returns compatibility scores with the broadcast shape of the observations.
    """

    def __init__(
        self,
        transmission_profile: InfectiousnessToTransmission,
        maximum_depth: int = 2,
        mc_samples: int = 10000,
        target: str | Scenario | Iterable[str | Scenario] = "ad(0)",
        mutation_process: str = "stochastic",
        rng: Generator | None = None,
    ) -> None:
        if maximum_depth < 0:
            raise ValueError("maximum_depth must be >= 0.")
        if mc_samples <= 0:
            raise ValueError("mc_samples must be > 0.")

        self.profile = transmission_profile
        self.maximum_depth = int(maximum_depth)
        self.sample_count = int(mc_samples)
        self.mutation_process = self._validate_mutation_process(mutation_process)
        self.random_generator = rng if rng is not None else transmission_profile.rng

        self.scenarios = self._enumerate_scenarios()
        self.scenarios_by_label = {scenario.label(): scenario for scenario in self.scenarios}
        self.target_labels = self._resolve_target_labels(target)
        self.target_label = self.target_labels[0] if len(self.target_labels) == 1 else None
        self._pairwise_models_by_target: dict[tuple[str, ...], PairwiseCompatibilityModel] = {}
        self.draws_by_scenario = self._precompute_draws()

    @property
    def draws_by_scenario(self) -> dict[str, dict[str, np.ndarray]]:
        return self._draws_by_scenario

    @draws_by_scenario.setter
    def draws_by_scenario(self, draws_by_scenario: dict[str, dict[str, np.ndarray]]) -> None:
        self._draws_by_scenario = draws_by_scenario
        self._pairwise_models_by_target.clear()

    @staticmethod
    def _validate_mutation_process(mutation_process: str) -> str:
        normalized_process = mutation_process.strip().lower()
        if normalized_process not in {"deterministic", "stochastic"}:
            raise ValueError("mutation_process must be either 'deterministic' or 'stochastic'.")
        return normalized_process

    @staticmethod
    def _normalize_target_label(target_label: str) -> str:
        return target_label.strip().lower()

    def _resolve_target_labels(
        self,
        target: str | Scenario | Iterable[str | Scenario],
    ) -> tuple[str, ...]:
        raw_targets: tuple[str | Scenario, ...]
        if isinstance(target, (str, Scenario)):
            raw_targets = (target,)
        else:
            raw_targets = tuple(target)

        if not raw_targets:
            raise ValueError("target must contain at least one scenario.")

        available_labels = ", ".join(self.scenarios_by_label)
        resolved_labels: set[str] = set()

        for raw_target in raw_targets:
            target_label = (
                raw_target.label() if isinstance(raw_target, Scenario) else str(raw_target)
            )
            normalized_label = self._normalize_target_label(target_label)

            if normalized_label not in self.scenarios_by_label:
                raise ValueError(
                    f"Unknown target scenario '{target_label}'. Available scenarios: {available_labels}."
                )
            resolved_labels.add(normalized_label)

        return tuple(label for label in self.scenarios_by_label if label in resolved_labels)

    def _enumerate_scenarios(self) -> list[Scenario]:
        scenarios: list[Scenario] = []

        for intermediates in range(self.maximum_depth + 1):
            scenarios.append(Scenario(kind="ad", intermediates=intermediates))

        for branch_to_i in range(self.maximum_depth + 1):
            for branch_to_j in range(self.maximum_depth + 1 - branch_to_i):
                scenarios.append(
                    Scenario(kind="ca", branch_to_i=branch_to_i, branch_to_j=branch_to_j)
                )

        return scenarios

    def _sum_generation_intervals(
        self,
        generations_per_path: int,
        sample_count: int,
    ) -> np.ndarray:
        if generations_per_path <= 0:
            return np.zeros(sample_count, dtype=float)

        generation_draws = self.profile.sample_generation_intervals(
            size=(generations_per_path, sample_count)
        )
        return np.asarray(generation_draws, dtype=float).sum(axis=0)

    def simulate_scenario_draws(
        self,
        scenario: Scenario,
        sample_count: int | None = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Simulate sample-time and branch-length draws for one scenario.

        Returns
        -------
        tuple[numpy.ndarray, numpy.ndarray]
            ``(time_draws, branch_draws)`` with one draw per Monte Carlo sample.
        """

        sample_count = self.sample_count if sample_count is None else int(sample_count)

        incubation_i = self.profile.sample_incubation_periods(sample_count)
        incubation_j = self.profile.sample_incubation_periods(sample_count)
        testing_delay_i = self.profile.sample_testing_delays(sample_count)
        testing_delay_j = self.profile.sample_testing_delays(sample_count)

        if scenario.kind == "ad":
            if scenario.intermediates is None:
                raise ValueError("Ancestor-descendant scenarios require intermediates.")
            elapsed_generations = self._sum_generation_intervals(
                generations_per_path=scenario.intermediates + 1,
                sample_count=sample_count,
            )
            time_draws = (
                elapsed_generations
                + incubation_j
                + testing_delay_j
                - incubation_i
                - testing_delay_i
            )
            branch_draws = time_draws
        elif scenario.kind == "ca":
            if scenario.branch_to_i is None or scenario.branch_to_j is None:
                raise ValueError("Common-ancestor scenarios require both branch depths.")
            path_to_i = self._sum_generation_intervals(
                generations_per_path=scenario.branch_to_i + 1,
                sample_count=sample_count,
            )
            path_to_j = self._sum_generation_intervals(
                generations_per_path=scenario.branch_to_j + 1,
                sample_count=sample_count,
            )
            time_draws = (
                path_to_j
                - path_to_i
                + incubation_j
                + testing_delay_j
                - incubation_i
                - testing_delay_i
            )
            branch_draws = (
                path_to_i
                + path_to_j
                + incubation_i
                + testing_delay_i
                + incubation_j
                + testing_delay_j
            )
        else:
            raise ValueError(f"Unknown scenario kind: {scenario.kind}")

        return np.asarray(time_draws, dtype=float), np.clip(
            np.asarray(branch_draws, dtype=float),
            a_min=0.0,
            a_max=np.inf,
        )

    def _simulate_genetic_draws(self, branch_draws: np.ndarray) -> np.ndarray:
        clock_rate_draws = np.asarray(
            self.profile.sample_clock_rate(size=branch_draws.shape),
            dtype=float,
        )
        expected_mutation_draws = np.clip(branch_draws * clock_rate_draws, a_min=0.0, a_max=np.inf)

        if self.mutation_process == "deterministic":
            return expected_mutation_draws

        return self.random_generator.poisson(expected_mutation_draws)

    def _precompute_draws(self) -> dict[str, dict[str, np.ndarray]]:
        draws_by_scenario: dict[str, dict[str, np.ndarray]] = {}

        for scenario in self.scenarios:
            label = scenario.label()
            time_draws, branch_draws = self.simulate_scenario_draws(scenario)
            genetic_draws = self._simulate_genetic_draws(branch_draws)

            draws_by_scenario[label] = {
                "time_draws": time_draws,
                "branch_draws": branch_draws,
                "genetic_draws": np.asarray(genetic_draws),
            }

        return draws_by_scenario

    @staticmethod
    def percentile_score(observed_value: float, simulated_values: np.ndarray) -> float:
        simulated_values = np.asarray(simulated_values, dtype=float)
        return float(np.mean(simulated_values <= observed_value))

    @staticmethod
    def compatibility_from_percentile(percentile: float) -> float:
        return float(1.0 - 2.0 * abs(percentile - 0.5))

    @classmethod
    def compatibility_score(
        cls,
        observed_value: float,
        simulated_values: np.ndarray,
    ) -> tuple[float, float]:
        percentile = cls.percentile_score(observed_value, simulated_values)
        compatibility = cls.compatibility_from_percentile(percentile)
        return percentile, compatibility

    @staticmethod
    def _validate_scalar_observation(observed_value: ArrayLike, name: str) -> float:
        scalar_value = np.asarray(observed_value, dtype=float)
        if scalar_value.ndim != 0:
            raise TypeError(
                f"score_pair() requires scalar {name}; use pairwise_model() or score_target() "
                "for arraylike inputs."
            )
        return float(scalar_value)

    def pairwise_model(
        self,
        target: str | Scenario | Iterable[str | Scenario] | None = None,
    ) -> PairwiseCompatibilityModel:
        """
        Return a cached vectorized scorer for a target subset.

        Parameters
        ----------
        target :
            Optional target subset. When omitted, the scorer uses the subset
            configured at construction time.
        """

        target_labels = (
            self.target_labels if target is None else self._resolve_target_labels(target)
        )
        pairwise_model = self._pairwise_models_by_target.get(target_labels)
        if pairwise_model is None:
            pairwise_model = PairwiseCompatibilityModel(self.draws_by_scenario, target_labels)
            self._pairwise_models_by_target[target_labels] = pairwise_model
        return pairwise_model

    def score_pair(
        self,
        sample_time_difference: float | None = None,
        genetic_distance: float | None = None,
    ) -> dict[str, Any]:
        """
        Compute compatibility scores for all scenarios for one sampled pair.

        Parameters
        ----------
        sample_time_difference :
            Observed difference in sampling times.
        genetic_distance :
            Observed consensus-level genetic distance.

        Returns
        -------
        dict
            Per-scenario percentile and compatibility summaries together with
            the summed compatibility of the configured target subset.
        """

        if sample_time_difference is None or genetic_distance is None:
            raise TypeError(
                "score_pair() requires both sample_time_difference and genetic_distance."
            )
        sample_time_difference = self._validate_scalar_observation(
            sample_time_difference,
            "sample_time_difference",
        )
        genetic_distance = self._validate_scalar_observation(
            genetic_distance,
            "genetic_distance",
        )

        scenario_scores = {}

        for label, draws in self.draws_by_scenario.items():
            time_percentile, time_compatibility = self.compatibility_score(
                observed_value=sample_time_difference,
                simulated_values=draws["time_draws"],
            )
            genetic_percentile, genetic_compatibility = self.compatibility_score(
                observed_value=genetic_distance,
                simulated_values=draws["genetic_draws"],
            )
            scenario_scores[label] = {
                "time_percentile": time_percentile,
                "time_compatibility": time_compatibility,
                "genetic_percentile": genetic_percentile,
                "genetic_compatibility": genetic_compatibility,
                "compatibility": time_compatibility * genetic_compatibility,
            }

        target_compatibility = float(
            self.pairwise_model()(sample_time_difference, genetic_distance)
        )

        return {
            "target": (
                self.target_labels[0] if self.target_label is not None else list(self.target_labels)
            ),
            "target_labels": list(self.target_labels),
            "target_compatibility": target_compatibility,
            "scenario_scores": scenario_scores,
        }

    def score_target(
        self,
        sample_time_difference: ArrayLike,
        genetic_distance: ArrayLike,
        target: str | Scenario | Iterable[str | Scenario] | None = None,
    ) -> float | np.ndarray:
        """
        Return the summed compatibility of a target subset.

        Scalar inputs return a scalar ``float``. Arraylike inputs return a
        ``numpy.ndarray`` with the broadcast shape of the observations.
        """

        result = self.pairwise_model(target).score(
            sample_time_difference=sample_time_difference,
            genetic_distance=genetic_distance,
        )
        return float(result) if result.ndim == 0 else result


__all__ = ["EpiLink", "PairwiseCompatibilityModel", "Scenario"]



from __future__ import annotations

import math
import sys
import unittest
from pathlib import Path

ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
if str(SRC) not in sys.path:
    sys.path.insert(0, str(SRC))

from epilink import NaturalHistoryParameters  # noqa: E402


class TestNaturalHistoryParameters(unittest.TestCase):
    def test_custom_parameters_update_derived_properties(self) -> None:
        p = NaturalHistoryParameters(
            incubation_shape=6.0,
            incubation_scale=0.75,
            latent_shape=2.5,
            symptomatic_rate=0.5,
            symptomatic_shape=2.0,
            transmission_rate_ratio=1.8,
        )

        self.assertTrue(math.isclose(p.presymptomatic_shape, 3.5))
        self.assertTrue(math.isclose(p.symptomatic_scale, 1.0))
        self.assertTrue(math.isclose(p.incubation_rate, 1.0 / 4.5))

    def test_to_dict_round_trips_through_constructor(self) -> None:
        original = NaturalHistoryParameters(
            incubation_shape=6.2,
            incubation_scale=0.8,
            latent_shape=2.0,
            symptomatic_rate=0.4,
            symptomatic_shape=1.3,
            transmission_rate_ratio=1.9,
            testing_delay_shape=2.5,
            testing_delay_scale=1.2,
            substitution_rate=1.2e-3,
            relaxation=0.15,
            genome_length=12345,
        )

        reconstructed = NaturalHistoryParameters(**original.to_dict())

        self.assertEqual(reconstructed, original)

    def test_default_derived_properties(self) -> None:
        p = NaturalHistoryParameters()

        self.assertTrue(math.isclose(p.presymptomatic_shape, p.incubation_shape - p.latent_shape))
        self.assertTrue(
            math.isclose(
                p.symptomatic_scale,
                1.0 / (p.symptomatic_shape * p.symptomatic_rate),
            )
        )
        self.assertTrue(
            math.isclose(
                p.incubation_rate,
                1.0 / (p.incubation_shape * p.incubation_scale),
            )
        )

    def test_transmission_metrics_match_documented_equations(self) -> None:
        p = NaturalHistoryParameters()

        expected_fraction = (
            p.transmission_rate_ratio * p.presymptomatic_shape * p.symptomatic_rate
        ) / (
            p.transmission_rate_ratio * p.presymptomatic_shape * p.symptomatic_rate
            + p.incubation_shape * p.incubation_rate
        )
        expected_normalisation = (p.incubation_shape * p.incubation_rate * p.symptomatic_rate) / (
            p.transmission_rate_ratio * p.presymptomatic_shape * p.symptomatic_rate
            + p.incubation_shape * p.incubation_rate
        )

        self.assertTrue(math.isclose(p.presymptomatic_transmission_fraction, expected_fraction))
        self.assertTrue(math.isclose(p.infectiousness_normalisation, expected_normalisation))
        self.assertGreaterEqual(p.presymptomatic_transmission_fraction, 0.0)
        self.assertLessEqual(p.presymptomatic_transmission_fraction, 1.0)

    def test_to_dict_returns_only_constructor_fields(self) -> None:
        p = NaturalHistoryParameters()
        data = p.to_dict()

        self.assertEqual(
            set(data.keys()),
            {
                "incubation_shape",
                "incubation_scale",
                "latent_shape",
                "symptomatic_rate",
                "symptomatic_shape",
                "transmission_rate_ratio",
                "testing_delay_shape",
                "testing_delay_scale",
                "substitution_rate",
                "relaxation",
                "genome_length",
            },
        )
        self.assertNotIn("presymptomatic_shape", data)
        self.assertEqual(data["testing_delay_shape"], 2.0)

    def test_bool_and_non_real_values_raise_type_error(self) -> None:
        with self.assertRaises(TypeError):
            NaturalHistoryParameters(symptomatic_rate=True)

        with self.assertRaises(TypeError):
            NaturalHistoryParameters(incubation_shape="5.807")

        with self.assertRaises(TypeError):
            NaturalHistoryParameters(relaxation=True)

    def test_non_positive_values_raise_value_error(self) -> None:
        non_positive_cases = (
            {"incubation_shape": 0.0},
            {"incubation_scale": -1.0},
            {"latent_shape": 0.0},
            {"symptomatic_rate": -0.1},
            {"symptomatic_shape": 0.0},
            {"transmission_rate_ratio": 0.0},
            {"testing_delay_shape": -2.0},
            {"testing_delay_scale": 0.0},
            {"substitution_rate": 0.0},
            {"genome_length": -1},
        )

        for kwargs in non_positive_cases:
            with self.subTest(kwargs=kwargs):
                with self.assertRaises(ValueError):
                    NaturalHistoryParameters(**kwargs)

    def test_relaxation_must_be_non_negative(self) -> None:
        NaturalHistoryParameters(relaxation=0.0)

        with self.assertRaises(ValueError):
            NaturalHistoryParameters(relaxation=-0.1)

    def test_latent_shape_must_be_less_than_incubation_shape(self) -> None:
        with self.assertRaises(ValueError):
            NaturalHistoryParameters(latent_shape=5.807)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from .epilink import EpiLink, PairwiseCompatibilityModel, Scenario
from .parameters import NaturalHistoryParameters
from .profiles import (
    BaseTransmissionProfile,
    InfectiousnessToTransmission,
    SymptomOnsetToTransmission,
)

__all__ = [
    "EpiLink",
    "PairwiseCompatibilityModel",
    "Scenario",
    "NaturalHistoryParameters",
    "BaseTransmissionProfile",
    "InfectiousnessToTransmission",
    "SymptomOnsetToTransmission",
]



from __future__ import annotations

import numpy as np
from numpy.random import Generator, default_rng
from scipy import stats
from scipy.integrate import cumulative_trapezoid

from ..model.parameters import NaturalHistoryParameters


class BaseTransmissionProfile:
    """Base class for transmission profile distributions.

    Parameters
    ----------
    parameters : NaturalHistoryParameters, optional
        Stage-duration and infectiousness parameters for the E/P/I model.
    grid_min_days : float, default=0.0
        Lower bound of the numerical sampling grid in days.
    grid_max_days : float, default=60.0
    rng : Generator, optional
        Random number generator to use for sampling.
    rng_seed : int, optional
        Seed used to initialize ``rng`` when no generator is supplied.
    grid_points : int, default=1024
        grid points used for numerical summaries and sampling.
    """

    def __init__(
        self,
        parameters: NaturalHistoryParameters | None = None,
        grid_min_days: float = 0.0,
        grid_max_days: float = 100.0,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        grid_points: int = 1024,
    ) -> None:

        if grid_max_days < grid_min_days:
            raise ValueError("grid_max_days must be greater than or equal to grid_min_days.")
        if grid_points < 2:
            raise ValueError("grid_points must be at least 2.")

        self.grid_min_days = float(grid_min_days)
        self.grid_max_days = float(grid_max_days)
        self.parameters = parameters or NaturalHistoryParameters()
        self.rng: Generator = rng if rng is not None else default_rng(rng_seed)
        self.grid_points = int(grid_points)

        self._sampling_grid: np.ndarray | None = None
        self._cdf_grid: np.ndarray | None = None

        # Frozen gamma distributions.
        param = self.parameters
        self.incubation = stats.gamma(a=param.incubation_shape, scale=param.incubation_scale)
        self.latent = stats.gamma(a=param.latent_shape, scale=param.incubation_scale)
        self.presymptomatic = stats.gamma(
            a=param.presymptomatic_shape, scale=param.incubation_scale
        )
        self.symptomatic = stats.gamma(a=param.symptomatic_shape, scale=param.symptomatic_scale)

    def _ensure_numerical_cdf(self) -> tuple[np.ndarray, np.ndarray]:
        if self._sampling_grid is None or self._cdf_grid is None:
            x = np.linspace(self.grid_min_days, self.grid_max_days, num=max(2, self.grid_points))
            y = np.clip(np.asarray(self.pdf(x), dtype=float), 0.0, np.inf)

            cdf = cumulative_trapezoid(y, x, initial=0)
            total = float(cdf[-1])

            if not np.isfinite(total) or total <= 0.0:
                cdf = (x - x[0]) / (x[-1] - x[0])
            else:
                cdf /= total
                cdf = np.maximum.accumulate(cdf)
                cdf[0] = 0.0
                cdf[-1] = 1.0

            self._sampling_grid = x
            self._cdf_grid = cdf

        return self._sampling_grid, self._cdf_grid

    def pdf(self, times_in_days: np.typing.ArrayLike) -> np.ndarray:
        """Evaluate the probability density function.

        Parameters
        ----------
        times_in_days : array_like
            Evaluation points in days.

        Returns
        -------
        numpy.ndarray
            Probability density values at ``times_in_days``.
        """

        raise NotImplementedError

    def cdf(self, times_in_days: np.typing.ArrayLike) -> np.ndarray:
        """Evaluate the cumulative distribution function on the numerical grid.

        Parameters
        ----------
        times_in_days : array_like
            Evaluation points in days.

        Returns
        -------
        numpy.ndarray
            Approximate cumulative probabilities clipped to ``[0, 1]``.
        """

        evaluation_points = np.asarray(times_in_days, dtype=float)
        x, cdf = self._ensure_numerical_cdf()
        values = np.interp(evaluation_points, x, cdf, left=0.0, right=1.0)
        return values[()] if evaluation_points.ndim == 0 else values

    def mean(self) -> float:
        """Estimate the mean on the configured numerical grid.

        Returns
        -------
        float
            Numerical estimate of the mean.
        """

        integration_grid = np.linspace(
            self.grid_min_days, self.grid_max_days, num=max(100, self.grid_points)
        )
        probability_density = np.clip(self.pdf(integration_grid), a_min=0.0, a_max=np.inf)
        total_mass = float(np.trapezoid(probability_density, integration_grid))
        if not np.isfinite(total_mass) or total_mass <= 0.0:
            return float("nan")
        return float(
            np.trapezoid(integration_grid * probability_density, integration_grid) / total_mass
        )

    def rvs(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        """Sample from the profile on the configured numerical grid.

        Parameters
        ----------
        size : int or tuple of int, default=1
            Output shape.

        Returns
        -------
        numpy.ndarray
            Random variates.
        """

        raise NotImplementedError

    def sample_latent_periods(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        """Sample latent-period durations.

        Parameters
        ----------
        size : int or tuple of int, default=1
            Output shape.

        Returns
        -------
        numpy.ndarray
            Latent-period durations in days.
        """

        return self.rng.gamma(
            shape=self.parameters.latent_shape,
            scale=self.parameters.incubation_scale,
            size=size,
        )

    def sample_presymptomatic_periods(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        """Sample presymptomatic-period durations.

        Parameters
        ----------
        size : int or tuple of int, default=1
            Output shape.

        Returns
        -------
        numpy.ndarray
            Presymptomatic-period durations in days.
        """

        return self.rng.gamma(
            shape=self.parameters.presymptomatic_shape,
            scale=self.parameters.incubation_scale,
            size=size,
        )

    def sample_incubation_periods(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        """Sample incubation periods.

        Parameters
        ----------
        size : int or tuple of int, default=1
            Output shape.

        Returns
        -------
        numpy.ndarray
            Incubation-period durations in days.
        """

        return self.sample_latent_periods(size) + self.sample_presymptomatic_periods(size)

    def sample_symptomatic_periods(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        """Sample symptomatic-period durations.

        Parameters
        ----------
        size : int or tuple of int, default=1
            Output shape.

        Returns
        -------
        numpy.ndarray
            Symptomatic-period durations in days.
        """

        return self.rng.gamma(
            shape=self.parameters.symptomatic_shape,
            scale=self.parameters.symptomatic_scale,
            size=size,
        )

    def sample_testing_delays(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        """Sample testing delays.

        Parameters
        ----------
        size : int or tuple of int, default=1
            Output shape.

        Returns
        -------
        numpy.ndarray
            Testing delays in days.
        """

        return self.rng.gamma(
            shape=self.parameters.testing_delay_shape,
            scale=self.parameters.testing_delay_scale,
            size=size,
        )

    def sample_clock_rate(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        """Sample substitution rates per day.

        Parameters
        ----------
        size : int or tuple of int, default=1
            Output shape.

        Returns
        -------
        numpy.ndarray
            Substitution rates in mutations per day.
        """
        per_site_per_year = self.rng.lognormal(
            np.log(self.parameters.substitution_rate), self.parameters.relaxation, size=size
        )
        return (per_site_per_year * self.parameters.genome_length) / 365.0

    def expected_mutations(
        self,
        times_in_days: np.typing.ArrayLike,
    ) -> np.ndarray:
        """Compute expected mutation counts for elapsed times.

        Parameters
        ----------
        times_in_days : array_like
            Times in days.
        size : int or tuple of int, default=1
            Shape of sampled rates when ``rates_per_day`` is None.

        Returns
        -------
        numpy.ndarray
            Expected mutation counts clipped to be non-negative.
        """

        times = np.asarray(times_in_days, dtype=float)
        clock_rate = self.sample_clock_rate(size=times.shape)
        return times * clock_rate

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"grid_min_days={self.grid_min_days}, "
            f"grid_max_days={self.grid_max_days}, "
            f"parameters={self.parameters})"
        )


class InfectiousnessToTransmission(BaseTransmissionProfile):
    r"""Distribution of time from the onset of infectiousness to transmission (:math:`y^*`).

    Density of the distribution is given by:

    .. math::

        f^*(y^*) =
        \begin{cases}
            0, & y^* < 0, \\
            C \left[ \alpha \left(1 - F_P(y^*)\right)
            + \int_0^{y^*} \left(1 - F_I(y^*-y_P)\right) f_P(y_P) \, d_{y_P} \right], & y^* \ge 0,
        \end{cases}

    where :math:`f_P` and :math:`F_P` are the presymptomatic PDF/CDF,
    :math:`F_I` is the symptomatic CDF, :math:`\alpha` is relative
    presymptomatic infectiousness, and :math:`C` is the normalisation
    constant from :class:`NaturalHistoryParameters`.

    Parameters
    ----------
    parameters : NaturalHistoryParameters, optional
        Stage-duration and infectiousness parameters for the E/P/I model.
    grid_min_days : float, default=0.0
        Lower bound of the numerical sampling grid in days.
    grid_max_days : float, default=60.0
        Upper bound of the numerical sampling grid in days.
    rng : Generator, optional
        Random number generator to use for sampling.
    rng_seed : int, optional
        Seed used to initialize ``rng`` when no generator is supplied.
    integration_grid_points : int, default=2048
         integration points used inside the PDF evaluation.
    sampling_grid_points : int, default=1024
         numerical grid points used for sampling from the profile.
    """

    def __init__(
        self,
        parameters: NaturalHistoryParameters | None = None,
        grid_min_days: float = 0.0,
        grid_max_days: float = 100.0,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        integration_grid_points: int = 2048,
        sampling_grid_points: int = 1024,
    ) -> None:

        super().__init__(
            grid_min_days=grid_min_days,
            grid_max_days=grid_max_days,
            parameters=parameters,
            rng=rng,
            rng_seed=rng_seed,
            grid_points=sampling_grid_points,
        )
        self.integration_grid_points = int(integration_grid_points)

    def sample_generation_intervals(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        """Sample generation intervals.

        Parameters
        ----------
        size : int or tuple of int, default=1
            Output shape.

        Returns
        -------
        numpy.ndarray
            Generation intervals in days.
        """

        return self.sample_latent_periods(size=size) + self.rvs(size=size)

    def pdf(self, times_in_days: np.typing.ArrayLike) -> np.ndarray:
        """Evaluate the infectiousness-to-transmission density.

        Parameters
        ----------
        times_in_days : array_like
            Evaluation points in days.

        Returns
        -------
        numpy.ndarray
            Probability density values at ``times_in_days``.
        """

        evaluation_points = np.asarray(times_in_days, dtype=float)
        flat_evaluation_points = evaluation_points.reshape(-1)
        probability_density = np.zeros_like(flat_evaluation_points)

        valid_indices = np.flatnonzero(flat_evaluation_points >= 0.0)
        if valid_indices.size == 0:
            if evaluation_points.ndim == 0:
                return probability_density.reshape(evaluation_points.shape)[()]
            return probability_density.reshape(evaluation_points.shape)

        param = self.parameters
        for index in valid_indices:
            toit_value = float(flat_evaluation_points[index])
            if toit_value == 0.0:
                symptomatic_contribution = 0.0
            else:
                presymptomatic_grid = np.linspace(
                    0.0,
                    toit_value,
                    num=max(2, self.integration_grid_points),
                )
                presymptomatic_density = self.presymptomatic.pdf(presymptomatic_grid)
                symptomatic_survival = 1.0 - self.symptomatic.cdf(toit_value - presymptomatic_grid)
                symptomatic_contribution = float(
                    np.trapezoid(symptomatic_survival * presymptomatic_density, presymptomatic_grid)
                )

            probability_density[index] = param.infectiousness_normalisation * (
                param.transmission_rate_ratio * (1.0 - self.presymptomatic.cdf(toit_value))
                + symptomatic_contribution
            )

        probability_density = probability_density.reshape(evaluation_points.shape)
        if evaluation_points.ndim == 0:
            return probability_density[()]
        return probability_density

    def rvs(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        """Sample transmission times from the configured numerical grid.

        Parameters
        ----------
        size : int or tuple of int, default=1
            Output shape.

        Returns
        -------
        numpy.ndarray
            Random variates from the profile.
        """

        sample_shape = (size,) if isinstance(size, int) else size
        x, cdf = self._ensure_numerical_cdf()
        u = self.rng.uniform(0.0, 1.0, size=sample_shape)
        return np.interp(u, cdf, x)


class SymptomOnsetToTransmission(BaseTransmissionProfile):
    r"""Distribution of time from the onset symptoms to transmission (:math:`x_{tost}`).

    Density of the distribution is given by:

    .. math::

        f_{tost}(x_{tost}) =
        \begin{cases}
            \alpha C \left(1 - F_P(-x_{tost})\right), & x_{tost} < 0, \\
            C \left(1 - F_I(x_{tost})\right), & x_{tost} \ge 0,
        \end{cases}

    where :math:`F_P` is the presymptomatic CDF, :math:`F_I` is the
    symptomatic CDF, :math:`\alpha` is relative presymptomatic
    infectiousness, and :math:`C` is the normalisation constant from
    :class:`NaturalHistoryParameters`.

    Parameters
    ----------
    parameters : NaturalHistoryParameters, optional
        Stage-duration and infectiousness parameters for the E/P/I model.
    grid_min_days : float, default=-30.0
        Lower bound of the numerical sampling grid in days.
    grid_max_days : float, default=30.0
        Upper bound of the numerical sampling grid in days.
    rng : Generator, optional
        Random number generator to use for sampling.
    rng_seed : int, optional
        Seed used to initialize ``rng`` when no generator is supplied.
    grid_points : int, default=2048
        numerical grid points used for summaries and sampling.
    """

    def __init__(
        self,
        parameters: NaturalHistoryParameters | None = None,
        grid_min_days: float = -30.0,
        grid_max_days: float = 30.0,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        grid_points: int = 2048,
    ) -> None:
        super().__init__(
            grid_min_days=grid_min_days,
            grid_max_days=grid_max_days,
            parameters=parameters,
            rng=rng,
            rng_seed=rng_seed,
            grid_points=grid_points,
        )

    def pdf(self, times_in_days: np.typing.ArrayLike) -> np.ndarray:
        r"""Evaluate the symptom-onset-to-transmission density.

        Parameters
        ----------
        times_in_days : array_like
            Evaluation points in days.

        Returns
        -------
        numpy.ndarray
            Probability density values at ``times_in_days``.
        """

        evaluation_points = np.asarray(times_in_days, dtype=float)
        parameters = self.parameters
        probability_density = np.where(
            evaluation_points < 0.0,
            parameters.transmission_rate_ratio
            * parameters.infectiousness_normalisation
            * (1.0 - self.presymptomatic.cdf(-evaluation_points)),
            parameters.infectiousness_normalisation
            * (1.0 - self.symptomatic.cdf(evaluation_points)),
        )
        probability_density = np.clip(probability_density, a_min=0.0, a_max=np.inf)
        if evaluation_points.ndim == 0:
            return probability_density[()]
        return probability_density

    def rvs(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        """Sample onset-to-transmission times from the configured grid.

        Parameters
        ----------
        size : int or tuple of int, default=1
            Output shape.

        Returns
        -------
        numpy.ndarray
            Random variates from the profile.
        """

        sample_shape = (size,) if isinstance(size, int) else size
        x, cdf = self._ensure_numerical_cdf()
        u = self.rng.uniform(0.0, 1.0, size=sample_shape)
        return np.interp(u, cdf, x)


__all__ = [
    "BaseTransmissionProfile",
    "InfectiousnessToTransmission",
    "SymptomOnsetToTransmission",
]



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

import networkx as nx
import numpy as np
import numpy.typing as npt
import pandas as pd

from ..model import InfectiousnessToTransmission
from .genome import PackedGenomicData


def simulate_epidemic_dates(
    transmission_profile: InfectiousnessToTransmission,
    tree: nx.DiGraph,
    fraction_sampled: float = 1.0,
) -> nx.DiGraph:
    """Populate a transmission tree with simulated epidemic dates.

    Parameters
    ----------
    transmission_profile : InfectiousnessToTransmission
        Transmission-time model used to sample latent, presymptomatic, and
        transmission intervals.
    tree : networkx.DiGraph
        Directed transmission tree.
    fraction_sampled : float, default=1.0
        Fraction of nodes marked as sampled.
        Shape parameter for the Gamma-distributed testing delay.

    Returns
    -------
    networkx.DiGraph
        Copy of ``tree`` with simulated epidemic annotations.
    """

    tree = tree.copy()
    rng = transmission_profile.rng

    num_nodes = tree.number_of_nodes()
    num_sampled = int(round(fraction_sampled * num_nodes))
    sampled_node_ids = set(rng.choice(list(tree.nodes()), size=num_sampled, replace=False))
    nx.set_node_attributes(tree, {node: (node in sampled_node_ids) for node in tree}, "sampled")

    def sample_stage_intervals() -> tuple[float, float, float]:
        latent = transmission_profile.sample_latent_periods().item()
        presymptomatic = transmission_profile.sample_presymptomatic_periods().item()
        testing = transmission_profile.sample_testing_delays().item()

        return latent, presymptomatic, testing

    roots = [n for n, d in tree.in_degree() if d == 0]

    for root in roots:
        exposure_date = int(rng.choice(range(30)))
        latent_period, presymptomatic_period, testing_delay = sample_stage_intervals()

        tree.nodes[root].update(
            {
                "exposure_date": exposure_date,
                "date_infectious": exposure_date + latent_period,
                "date_symptom_onset": (exposure_date + latent_period + presymptomatic_period),
                "sample_date": (
                    exposure_date + latent_period + presymptomatic_period + testing_delay
                ),
                "seed": True,
            }
        )

        for parent, child in nx.dfs_edges(tree, source=root):
            transmission_time = transmission_profile.rvs().item()
            latent_period, presymptomatic_period, testing_delay = sample_stage_intervals()
            parent_infectious_date = tree.nodes[parent]["date_infectious"]

            child_exposure_date = parent_infectious_date + transmission_time
            child_infectious_date = child_exposure_date + latent_period
            child_symptom_onset_date = child_infectious_date + presymptomatic_period
            child_sample_date = child_symptom_onset_date + testing_delay

            tree.nodes[child].update(
                {
                    "exposure_date": child_exposure_date,
                    "date_infectious": child_infectious_date,
                    "date_symptom_onset": child_symptom_onset_date,
                    "sample_date": child_sample_date,
                    "seed": False,
                }
            )

    return tree


def simulate_genomic_sequences(
    transmission_profile: InfectiousnessToTransmission,
    tree: nx.DiGraph,
    genome_length: int = 1000,
    return_raw: bool = False,
) -> dict[str, Any]:
    """Simulate genomic evolution along a transmission tree.

    Parameters
    ----------
    transmission_profile : InfectiousnessToTransmission
        Transmission-time model used to sample latent, presymptomatic, and
        transmission intervals.
    tree : networkx.DiGraph
        Directed transmission tree with epidemic-date annotations.
    genome_length : int, default=1000
        Length of the simulated genome sequences.
    return_raw : bool, default=False
        If ``True``, include the unpacked integer sequence matrices in the
        returned dictionary.

    Returns
    -------
    dict
        Dictionary with packed genomic data and, optionally, raw sequences.
    """

    base_lookup = {0: "A", 1: "C", 2: "G", 3: "T"}
    num_bases = len(base_lookup)
    nodes = list(tree.nodes())
    node_to_index = {node: index for index, node in enumerate(nodes)}
    num_nodes = len(nodes)

    rng = transmission_profile.rng

    linear_sequences = np.zeros((num_nodes, genome_length), dtype=np.int8)
    poisson_sequences = np.zeros((num_nodes, genome_length), dtype=np.int8)
    reference_sequence = rng.integers(0, num_bases, size=genome_length, dtype=np.int8)

    def mutate_sequence(
        sequence: npt.NDArray[np.int8],
        num_mutations: int,
    ) -> npt.NDArray[np.int8]:
        num_mutations = min(int(num_mutations), genome_length)
        if num_mutations <= 0:
            return sequence.copy()

        mutated_sequence = sequence.copy()
        mutation_sites = np.asarray(
            rng.choice(genome_length, size=num_mutations, replace=False),
            dtype=np.intp,
        )
        mutation_offsets = rng.integers(1, num_bases, size=num_mutations, dtype=np.int8)
        mutated_sequence[mutation_sites] = (
            mutated_sequence[mutation_sites] + mutation_offsets
        ) % num_bases
        return mutated_sequence

    roots = [node for node, degree in tree.in_degree(tree.nodes) if degree == 0]
    for root in roots:
        root_index = node_to_index[root]
        max_root_drift = min(35, genome_length + 1)
        root_drift_mutations = int(rng.integers(1, max_root_drift))
        root_sequence = mutate_sequence(reference_sequence, root_drift_mutations)
        linear_sequences[root_index] = root_sequence
        poisson_sequences[root_index] = root_sequence

    for root in roots:
        for parent, child in nx.dfs_edges(tree, source=root):
            parent_index = node_to_index[parent]
            child_index = node_to_index[child]

            try:
                transmission_date = tree.nodes[child]["exposure_date"]
                parent_sample_date = tree.nodes[parent]["sample_date"]
                child_sample_date = tree.nodes[child]["sample_date"]
            except KeyError as error:
                raise ValueError("Missing dates in tree. Run epidemic simulation first.") from error

            branch_length_days = abs(parent_sample_date - transmission_date) + abs(
                child_sample_date - transmission_date
            )

            expected_mutations = transmission_profile.expected_mutations(branch_length_days)

            num_linear_mutations = int(round(expected_mutations.item()))
            linear_sequences[child_index] = mutate_sequence(
                linear_sequences[parent_index],
                num_linear_mutations,
            )

            num_poisson_mutations = int(rng.poisson(expected_mutations))
            poisson_sequences[child_index] = mutate_sequence(
                poisson_sequences[parent_index],
                num_poisson_mutations,
            )

    packed_sequences = {
        "linear": PackedGenomicData(
            linear_sequences,
            genome_length,
            node_to_index,
            base_lookup,
        ),
        "poisson": PackedGenomicData(
            poisson_sequences,
            genome_length,
            node_to_index,
            base_lookup,
        ),
    }
    raw_sequences = {"linear": linear_sequences, "poisson": poisson_sequences}

    return {"packed": packed_sequences, "raw": raw_sequences if return_raw else None}


def build_pairwise_case_table(
    packed_genomic_data: Mapping[str, PackedGenomicData], tree: nx.DiGraph
) -> pd.DataFrame:
    """Generate a long-format pairwise distance table.

    Parameters
    ----------
    packed_genomic_data : mapping
        Mapping containing packed genomic data for the ``"linear"`` and
        ``"poisson"`` simulation outputs.
    tree : networkx.DiGraph
        Directed transmission tree with epidemic annotations.

    Returns
    -------
    pandas.DataFrame
        Pairwise table of genetic distances, temporal distances, and simple
        topological relationships.
    """

    packed_linear = packed_genomic_data["linear"]
    packed_poisson = packed_genomic_data["poisson"]
    node_map = packed_linear.node_to_idx
    n_nodes = packed_linear.n_seqs
    idx_to_node = {v: k for k, v in node_map.items()}

    mat_linear = packed_linear.compute_hamming_distances()
    mat_poisson = packed_poisson.compute_hamming_distances()

    sample_dates = np.zeros(n_nodes)
    for node, idx in node_map.items():
        sample_dates[idx] = tree.nodes[node].get("sample_date", np.nan)

    diff_matrix = sample_dates[:, np.newaxis] - sample_dates
    mat_temporal = np.abs(diff_matrix).round()

    mat_related = np.zeros((n_nodes, n_nodes), dtype=bool)

    for u, v in tree.edges():
        if u in node_map and v in node_map:
            i, j = node_map[u], node_map[v]
            mat_related[i, j] = True
            mat_related[j, i] = True

    for node in tree.nodes():
        children = list(tree.successors(node))
        if len(children) > 1:
            child_indices = [node_map[c] for c in children if c in node_map]
            if child_indices:
                grid_x, grid_y = np.meshgrid(child_indices, child_indices)
                mat_related[grid_x, grid_y] = True

    rows, cols = np.triu_indices(n_nodes, k=1)
    id_array = np.array([idx_to_node[i] for i in range(n_nodes)])
    sampled_status = np.array([tree.nodes[node].get("sampled", False) for node in id_array])

    return pd.DataFrame(
        {
            "CaseA": id_array[rows],
            "CaseB": id_array[cols],
            "IsRelated": mat_related[rows, cols],
            "BothSampled": sampled_status[rows] & sampled_status[cols],
            "DeterministicDistance": mat_linear[rows, cols],
            "StochasticDistance": mat_poisson[rows, cols],
            "SamplingDateDistanceDays": mat_temporal[rows, cols],
        }
    )


__all__ = ["simulate_genomic_sequences", "build_pairwise_case_table", "simulate_epidemic_dates"]



from __future__ import annotations

import sys
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

import networkx as nx
import numpy as np

ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
if str(SRC) not in sys.path:
    sys.path.insert(0, str(SRC))

from epilink import (  # noqa: E402
    PackedGenomicData,
    SequencePacker64,
    build_pairwise_case_table,
    simulate_epidemic_dates,
    simulate_genomic_sequences,
)

BASE_MAP = {0: "A", 1: "C", 2: "G", 3: "T"}


class DeterministicSimulationProfile:
    def __init__(
        self,
        *,
        latent: float = 2.0,
        presymptomatic: float = 3.0,
        testing: float = 4.0,
        transmission: float = 5.0,
        expected_mutation_scale: float = 0.0,
        rng_seed: int = 123,
    ) -> None:
        self.latent = latent
        self.presymptomatic = presymptomatic
        self.testing = testing
        self.transmission = transmission
        self.expected_mutation_scale = expected_mutation_scale
        self.rng = np.random.default_rng(rng_seed)

    @staticmethod
    def _full(
        size: int | tuple[int, ...] = 1,
        value: float = 0.0,
    ) -> np.ndarray:
        sample_shape = (size,) if isinstance(size, int) else size
        return np.full(sample_shape, value, dtype=float)

    def sample_latent_periods(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        return self._full(size, self.latent)

    def sample_presymptomatic_periods(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        return self._full(size, self.presymptomatic)

    def sample_testing_delays(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        return self._full(size, self.testing)

    def rvs(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        return self._full(size, self.transmission)

    def expected_mutations(self, branch_length_days: float | np.ndarray) -> np.ndarray:
        return np.asarray(branch_length_days, dtype=float) * self.expected_mutation_scale


class TestSequencePacker64(unittest.TestCase):
    def test_pack_u64_and_hamming64_match_direct_hamming_distances(self) -> None:
        sequence_a = np.tile(np.array([0, 1, 2, 3], dtype=np.int8), 9)
        sequence_b = sequence_a.copy()
        sequence_b[[0, 17, 35]] = np.array([1, 2, 0], dtype=np.int8)
        sequence_c = np.full(36, 3, dtype=np.int8)
        sequences = np.vstack([sequence_a, sequence_b, sequence_c])

        packed = SequencePacker64.pack_u64(sequences)
        distances = SequencePacker64.hamming64(packed)
        expected = np.count_nonzero(
            sequences[:, np.newaxis, :] != sequences[np.newaxis, :, :],
            axis=2,
        ).astype(np.int32)

        self.assertEqual(packed.shape, (3, 2))
        self.assertEqual(packed.dtype, np.uint64)
        np.testing.assert_array_equal(distances, expected)

    def test_hamming64_matches_direct_distances_across_multiple_blocks(self) -> None:
        rng = np.random.default_rng(7)
        sequences = rng.integers(0, 4, size=(6, 97), dtype=np.int8)

        packed = SequencePacker64.pack_u64(sequences)
        distances = SequencePacker64.hamming64(packed, block_size=2)
        expected = np.count_nonzero(
            sequences[:, np.newaxis, :] != sequences[np.newaxis, :, :],
            axis=2,
        ).astype(np.int32)

        np.testing.assert_array_equal(distances, expected)


class TestPackedGenomicData(unittest.TestCase):
    def test_validates_input_matrix_dtype_and_value_range(self) -> None:
        with self.assertRaises(TypeError):
            PackedGenomicData(
                np.array([[0.0, 1.0]], dtype=float),
                original_length=2,
                node_map={"case-1": 0},
                base_map=BASE_MAP,
            )

        with self.assertRaises(ValueError):
            PackedGenomicData(
                np.array([[0, 4]], dtype=np.int8),
                original_length=2,
                node_map={"case-1": 0},
                base_map=BASE_MAP,
            )

    def test_compute_hamming_distances_handles_non_contiguous_packed_storage(self) -> None:
        base_pattern = np.tile(np.array([0, 1, 2, 3], dtype=np.int8), 10)
        sequences = np.vstack(
            [
                base_pattern,
                np.roll(base_pattern, 1),
                np.roll(base_pattern, 5),
            ]
        )
        packed = PackedGenomicData(
            np.asfortranarray(sequences),
            original_length=sequences.shape[1],
            node_map={"A": 0, "B": 1, "C": 2},
            base_map=BASE_MAP,
        )
        packed.packed_u64 = packed.packed_u64[:, ::-1]
        expected = np.count_nonzero(
            sequences[:, np.newaxis, :] != sequences[np.newaxis, :, :],
            axis=2,
        ).astype(np.int32)

        self.assertFalse(packed.packed_u64.flags["C_CONTIGUOUS"])
        np.testing.assert_array_equal(packed.compute_hamming_distances(block_size=1), expected)

    def test_write_fasta_outputs_wrapped_sequences_with_node_headers(self) -> None:
        sequence_a = np.concatenate(
            [
                np.zeros(100, dtype=np.int8),
                np.array([1, 2, 3, 0, 1], dtype=np.int8),
            ]
        )
        sequence_b = np.full(105, 3, dtype=np.int8)
        packed = PackedGenomicData(
            np.vstack([sequence_a, sequence_b]),
            original_length=105,
            node_map={"case-1": 0, "case-2": 1},
            base_map=BASE_MAP,
        )

        with TemporaryDirectory() as temp_dir:
            fasta_path = Path(temp_dir) / "simulated.fasta"
            packed.write_fasta(str(fasta_path))
            lines = fasta_path.read_text().splitlines()

        self.assertEqual(lines[0], ">case-1")
        self.assertEqual(lines[1], "A" * 100)
        self.assertEqual(lines[2], "CGTAC")
        self.assertEqual(lines[3], ">case-2")
        self.assertEqual(lines[4], "T" * 100)
        self.assertEqual(lines[5], "T" * 5)


class TestSimulationHelpers(unittest.TestCase):
    def test_simulate_epidemic_dates_adds_dates_without_mutating_input_tree(self) -> None:
        tree = nx.DiGraph([("root", "child")])
        tree.add_node("other-root")
        profile = DeterministicSimulationProfile()

        simulated = simulate_epidemic_dates(profile, tree, fraction_sampled=0.5)

        self.assertNotIn("sampled", tree.nodes["root"])
        self.assertEqual(sum(simulated.nodes[node]["sampled"] for node in simulated.nodes), 2)

        root = simulated.nodes["root"]
        child = simulated.nodes["child"]
        other_root = simulated.nodes["other-root"]

        self.assertTrue(root["seed"])
        self.assertFalse(child["seed"])
        self.assertTrue(other_root["seed"])
        self.assertGreaterEqual(root["exposure_date"], 0)
        self.assertLess(root["exposure_date"], 30)
        self.assertEqual(root["date_infectious"] - root["exposure_date"], 2.0)
        self.assertEqual(root["date_symptom_onset"] - root["date_infectious"], 3.0)
        self.assertEqual(root["sample_date"] - root["date_symptom_onset"], 4.0)
        self.assertEqual(child["exposure_date"], root["date_infectious"] + 5.0)
        self.assertEqual(child["date_infectious"] - child["exposure_date"], 2.0)
        self.assertEqual(child["date_symptom_onset"] - child["date_infectious"], 3.0)
        self.assertEqual(child["sample_date"] - child["date_symptom_onset"], 4.0)

    def test_simulate_genomic_sequences_requires_epidemic_dates(self) -> None:
        tree = nx.DiGraph([("root", "child")])

        with self.assertRaises(ValueError):
            simulate_genomic_sequences(
                DeterministicSimulationProfile(),
                tree,
                genome_length=8,
            )

    def test_simulate_genomic_sequences_returns_raw_and_packed_outputs(self) -> None:
        tree = nx.DiGraph([("root", "child")])
        tree.nodes["root"].update(exposure_date=0.0, sample_date=6.0)
        tree.nodes["child"].update(exposure_date=2.0, sample_date=8.0)

        result = simulate_genomic_sequences(
            DeterministicSimulationProfile(expected_mutation_scale=0.0, rng_seed=9),
            tree,
            genome_length=12,
            return_raw=True,
        )

        self.assertEqual(set(result["packed"]), {"linear", "poisson"})
        self.assertEqual(set(result["raw"]), {"linear", "poisson"})
        self.assertEqual(result["packed"]["linear"].node_to_idx, {"root": 0, "child": 1})
        self.assertEqual(result["packed"]["linear"].original_length, 12)
        self.assertEqual(result["raw"]["linear"].shape, (2, 12))
        self.assertTrue(np.all((result["raw"]["linear"] >= 0) & (result["raw"]["linear"] <= 3)))
        np.testing.assert_array_equal(result["raw"]["linear"][0], result["raw"]["linear"][1])
        np.testing.assert_array_equal(result["raw"]["poisson"][0], result["raw"]["poisson"][1])
        np.testing.assert_array_equal(
            result["packed"]["linear"].compute_hamming_distances(),
            np.zeros((2, 2), dtype=np.int32),
        )

        without_raw = simulate_genomic_sequences(
            DeterministicSimulationProfile(expected_mutation_scale=0.0, rng_seed=9),
            tree,
            genome_length=12,
            return_raw=False,
        )
        self.assertIsNone(without_raw["raw"])

    def test_simulate_genomic_sequences_linear_model_mutates_selected_sites(self) -> None:
        tree = nx.DiGraph([("root", "child")])
        tree.nodes["root"].update(exposure_date=0.0, sample_date=6.0)
        tree.nodes["child"].update(exposure_date=2.0, sample_date=8.0)

        result = simulate_genomic_sequences(
            DeterministicSimulationProfile(expected_mutation_scale=0.5, rng_seed=9),
            tree,
            genome_length=4,
            return_raw=True,
        )

        linear_sequences = result["raw"]["linear"]
        self.assertEqual(np.count_nonzero(linear_sequences[0] != linear_sequences[1]), 4)
        self.assertTrue(np.all((linear_sequences >= 0) & (linear_sequences <= 3)))

    def test_build_pairwise_case_table_reports_relationships_sampling_and_distances(self) -> None:
        tree = nx.DiGraph([("A", "B"), ("A", "C")])
        tree.add_node("D")
        nx.set_node_attributes(
            tree,
            {
                "A": {"sample_date": 10.0, "sampled": True},
                "B": {"sample_date": 12.0, "sampled": True},
                "C": {"sample_date": 17.0, "sampled": False},
                "D": {"sample_date": 20.0, "sampled": True},
            },
        )

        packed = {
            "linear": PackedGenomicData(
                np.array(
                    [
                        [0, 0, 0, 0],
                        [0, 1, 0, 0],
                        [2, 0, 0, 0],
                        [3, 3, 3, 3],
                    ],
                    dtype=np.int8,
                ),
                original_length=4,
                node_map={"A": 0, "B": 1, "C": 2, "D": 3},
                base_map=BASE_MAP,
            ),
            "poisson": PackedGenomicData(
                np.array(
                    [
                        [0, 0, 0, 0],
                        [0, 1, 1, 0],
                        [2, 2, 0, 0],
                        [3, 3, 3, 0],
                    ],
                    dtype=np.int8,
                ),
                original_length=4,
                node_map={"A": 0, "B": 1, "C": 2, "D": 3},
                base_map=BASE_MAP,
            ),
        }

        table = build_pairwise_case_table(packed, tree)
        rows = {(row.CaseA, row.CaseB): row for row in table.itertuples(index=False)}

        self.assertEqual(len(table), 6)
        self.assertTrue(rows[("A", "B")].IsRelated)
        self.assertTrue(rows[("A", "B")].BothSampled)
        self.assertEqual(rows[("A", "B")].DeterministicDistance, 1)
        self.assertEqual(rows[("A", "B")].StochasticDistance, 2)
        self.assertEqual(rows[("A", "B")].SamplingDateDistanceDays, 2.0)

        self.assertTrue(rows[("B", "C")].IsRelated)
        self.assertFalse(rows[("B", "C")].BothSampled)
        self.assertEqual(rows[("B", "C")].DeterministicDistance, 2)
        self.assertEqual(rows[("B", "C")].StochasticDistance, 3)
        self.assertEqual(rows[("B", "C")].SamplingDateDistanceDays, 5.0)

        self.assertFalse(rows[("A", "D")].IsRelated)
        self.assertTrue(rows[("A", "D")].BothSampled)
        self.assertEqual(rows[("A", "D")].DeterministicDistance, 4)
        self.assertEqual(rows[("A", "D")].StochasticDistance, 3)
        self.assertEqual(rows[("A", "D")].SamplingDateDistanceDays, 10.0)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from .model import (
    BaseTransmissionProfile,
    EpiLink,
    InfectiousnessToTransmission,
    NaturalHistoryParameters,
    PairwiseCompatibilityModel,
    Scenario,
    SymptomOnsetToTransmission,
)
from .simulation import (
    PackedGenomicData,
    SequencePacker64,
    build_pairwise_case_table,
    simulate_epidemic_dates,
    simulate_genomic_sequences,
)

__version__ = "0.1.3"

__all__ = [
    "EpiLink",
    "PairwiseCompatibilityModel",
    "Scenario",
    "NaturalHistoryParameters",
    "BaseTransmissionProfile",
    "InfectiousnessToTransmission",
    "SymptomOnsetToTransmission",
    "SequencePacker64",
    "PackedGenomicData",
    "simulate_epidemic_dates",
    "simulate_genomic_sequences",
    "build_pairwise_case_table",
]



from __future__ import annotations

import sys
import unittest
from pathlib import Path

import numpy as np

ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
if str(SRC) not in sys.path:
    sys.path.insert(0, str(SRC))

from epilink import EpiLink, Scenario  # noqa: E402


class CountingProfile:
    def __init__(self) -> None:
        self.rng = np.random.default_rng(1234)
        self.call_counts = {
            "sample_incubation_periods": 0,
            "sample_testing_delays": 0,
            "sample_generation_intervals": 0,
            "sample_clock_rate": 0,
        }

    @staticmethod
    def _sequence(size: int | tuple[int, ...], start: float, step: float = 0.1) -> np.ndarray:
        shape = size if isinstance(size, tuple) else (size,)
        total = int(np.prod(shape))
        values = start + step * np.arange(total, dtype=float)
        return values.reshape(shape)

    def sample_incubation_periods(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        self.call_counts["sample_incubation_periods"] += 1
        return self._sequence(size, start=1.0)

    def sample_testing_delays(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        self.call_counts["sample_testing_delays"] += 1
        return self._sequence(size, start=0.5, step=0.05)

    def sample_generation_intervals(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        self.call_counts["sample_generation_intervals"] += 1
        return self._sequence(size, start=2.0, step=0.2)

    def sample_clock_rate(self, size: int | tuple[int, ...] = 1) -> np.ndarray:
        self.call_counts["sample_clock_rate"] += 1
        return self._sequence(size, start=0.2, step=0.01)


class TestEpiLink(unittest.TestCase):
    @staticmethod
    def _manual_draws() -> dict[str, dict[str, np.ndarray]]:
        return {
            "ad(0)": {
                "time_draws": np.array([1.0, 2.0, 3.0, 4.0]),
                "branch_draws": np.array([1.0, 1.0, 1.0, 1.0]),
                "genetic_draws": np.array([0.0, 1.0, 2.0, 3.0]),
            },
            "ca(0,0)": {
                "time_draws": np.array([2.0, 3.0, 4.0, 5.0]),
                "branch_draws": np.array([2.0, 2.0, 2.0, 2.0]),
                "genetic_draws": np.array([1.0, 2.0, 3.0, 4.0]),
            },
        }

    def test_score_pair_uses_cached_draws(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=1,
            mc_samples=8,
            target="ad(0)",
            mutation_process="deterministic",
        )

        call_counts_after_init = dict(profile.call_counts)

        first_result = model.score_pair(sample_time_difference=2.0, genetic_distance=1.0)
        second_result = model.score_pair(sample_time_difference=3.0, genetic_distance=2.0)

        self.assertEqual(profile.call_counts, call_counts_after_init)
        self.assertEqual(first_result["target"], "ad(0)")
        self.assertIn("scenario_scores", second_result)

    def test_score_pair_returns_target_compatibility_and_per_scenario_scores(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=0,
            mc_samples=4,
            target="ad(0)",
            mutation_process="deterministic",
        )

        model.draws_by_scenario = self._manual_draws()
        model.draws_by_scenario["ca(0,0)"]["time_draws"] = np.array([3.0, 4.0, 5.0, 6.0])
        model.draws_by_scenario["ca(0,0)"]["genetic_draws"] = np.array([4.0, 5.0, 6.0, 7.0])

        result = model.score_pair(sample_time_difference=2.5, genetic_distance=1.5)

        self.assertAlmostEqual(result["scenario_scores"]["ad(0)"]["time_percentile"], 0.5)
        self.assertAlmostEqual(result["scenario_scores"]["ad(0)"]["genetic_percentile"], 0.5)
        self.assertAlmostEqual(result["scenario_scores"]["ad(0)"]["compatibility"], 1.0)
        self.assertAlmostEqual(result["scenario_scores"]["ca(0,0)"]["compatibility"], 0.0)
        self.assertAlmostEqual(result["target_compatibility"], 1.0)
        self.assertAlmostEqual(model.score_target(2.5, 1.5), 1.0)

    def test_score_pair_sums_compatibility_across_target_subset(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=0,
            mc_samples=4,
            target=["ca(0,0)", "ad(0)", "ad(0)"],
            mutation_process="deterministic",
        )

        model.draws_by_scenario = self._manual_draws()

        result = model.score_pair(sample_time_difference=2.5, genetic_distance=1.5)

        self.assertEqual(model.target_labels, ("ad(0)", "ca(0,0)"))
        self.assertIsNone(model.target_label)
        self.assertEqual(result["target"], ["ad(0)", "ca(0,0)"])
        self.assertEqual(result["target_labels"], ["ad(0)", "ca(0,0)"])
        self.assertAlmostEqual(result["scenario_scores"]["ad(0)"]["compatibility"], 1.0)
        self.assertAlmostEqual(result["scenario_scores"]["ca(0,0)"]["compatibility"], 0.25)
        self.assertAlmostEqual(result["target_compatibility"], 1.25)
        self.assertAlmostEqual(model.score_target(2.5, 1.5), 1.25)

    def test_pairwise_model_accepts_arraylike_inputs_and_returns_broadcast_shape(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=0,
            mc_samples=4,
            target=["ad(0)", "ca(0,0)"],
            mutation_process="deterministic",
        )
        model.draws_by_scenario = self._manual_draws()

        pairwise_model = model.pairwise_model()
        sample_time_difference = np.array([[2.5], [4.5]])
        genetic_distance = np.array([[1.5, 2.5, 3.5]])

        scores = pairwise_model(sample_time_difference, genetic_distance)
        broadcast_time, broadcast_genetic = np.broadcast_arrays(
            sample_time_difference,
            genetic_distance,
        )
        expected = np.empty_like(broadcast_time, dtype=float)

        for index in np.ndindex(expected.shape):
            expected[index] = model.score_pair(
                sample_time_difference=float(broadcast_time[index]),
                genetic_distance=float(broadcast_genetic[index]),
            )["target_compatibility"]

        np.testing.assert_allclose(scores, expected)
        self.assertEqual(scores.shape, expected.shape)
        np.testing.assert_allclose(
            model.score_target(sample_time_difference, genetic_distance),
            expected,
        )

    def test_pairwise_model_caches_equivalent_target_subsets(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=0,
            mc_samples=4,
            mutation_process="deterministic",
        )

        first_model = model.pairwise_model(["ca(0,0)", "ad(0)"])
        second_model = model.pairwise_model(["ad(0)", "ca(0,0)", "ad(0)"])

        self.assertIs(first_model, second_model)
        self.assertEqual(first_model.target_labels, ("ad(0)", "ca(0,0)"))

    def test_draws_by_scenario_setter_invalidates_cached_pairwise_models(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=0,
            mc_samples=4,
            mutation_process="deterministic",
        )

        first_model = model.pairwise_model(["ad(0)"])
        replacement_draws = {
            label: {name: values.copy() for name, values in payload.items()}
            for label, payload in self._manual_draws().items()
        }

        model.draws_by_scenario = replacement_draws
        second_model = model.pairwise_model(["ad(0)"])

        self.assertIsNot(first_model, second_model)

    def test_target_accepts_scenario_objects_and_preserves_canonical_order(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=0,
            mc_samples=4,
            target=[
                Scenario(kind="ca", branch_to_i=0, branch_to_j=0),
                Scenario(kind="ad", intermediates=0),
            ],
            mutation_process="deterministic",
        )

        self.assertEqual(model.target_labels, ("ad(0)", "ca(0,0)"))
        self.assertIsNone(model.target_label)

    def test_score_target_returns_scalar_float_for_scalar_inputs(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=0,
            mc_samples=4,
            mutation_process="deterministic",
        )
        model.draws_by_scenario = self._manual_draws()

        score = model.score_target(2.5, 1.5, target="ad(0)")

        self.assertIsInstance(score, float)
        self.assertAlmostEqual(score, 1.0)

    def test_deterministic_mutation_process_uses_expected_mutation_counts(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=0,
            mc_samples=4,
            mutation_process="deterministic",
        )

        draws = model.draws_by_scenario["ad(0)"]
        expected_branch_draws = np.array([2.0, 2.2, 2.4, 2.6])
        expected_clock_rate_draws = np.array([0.2, 0.21, 0.22, 0.23])

        np.testing.assert_allclose(draws["branch_draws"], expected_branch_draws)
        np.testing.assert_allclose(
            draws["genetic_draws"],
            expected_branch_draws * expected_clock_rate_draws,
        )
        self.assertTrue(np.issubdtype(draws["genetic_draws"].dtype, np.floating))

    def test_stochastic_mutation_process_precomputes_integer_genetic_draws(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=0,
            mc_samples=16,
            mutation_process="stochastic",
        )

        genetic_draws = model.draws_by_scenario["ad(0)"]["genetic_draws"]

        self.assertTrue(np.issubdtype(genetic_draws.dtype, np.integer))

    def test_simulate_common_ancestor_draws_match_manual_expectation(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=1,
            mc_samples=3,
            mutation_process="deterministic",
        )

        time_draws, branch_draws = model.simulate_scenario_draws(
            Scenario(kind="ca", branch_to_i=0, branch_to_j=1),
            sample_count=3,
        )

        np.testing.assert_allclose(time_draws, np.array([2.6, 2.8, 3.0]))
        np.testing.assert_allclose(branch_draws, np.array([9.6, 10.5, 11.4]))

    def test_legacy_target_labels_raise_value_error(self) -> None:
        profile = CountingProfile()

        with self.assertRaises(ValueError):
            EpiLink(
                transmission_profile=profile,
                maximum_depth=0,
                mc_samples=4,
                target="HCA(0,0)",
                mutation_process="deterministic",
            )

    def test_legacy_score_pair_keywords_raise_type_error(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=0,
            mc_samples=4,
            mutation_process="deterministic",
        )

        with self.assertRaises(TypeError):
            model.score_pair(t_ij=2.5, g_ij=1.5)

    def test_invalid_configuration_raises_value_error(self) -> None:
        profile = CountingProfile()

        with self.subTest(parameter="maximum_depth"):
            with self.assertRaises(ValueError):
                EpiLink(
                    transmission_profile=profile,
                    maximum_depth=-1,
                    mc_samples=4,
                    mutation_process="deterministic",
                )

        with self.subTest(parameter="mc_samples"):
            with self.assertRaises(ValueError):
                EpiLink(
                    transmission_profile=profile,
                    maximum_depth=0,
                    mc_samples=0,
                    mutation_process="deterministic",
                )

        with self.subTest(parameter="mutation_process"):
            with self.assertRaises(ValueError):
                EpiLink(
                    transmission_profile=profile,
                    maximum_depth=0,
                    mc_samples=4,
                    mutation_process="not-a-process",
                )

    def test_score_pair_rejects_arraylike_inputs(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=0,
            mc_samples=4,
            mutation_process="deterministic",
        )

        with self.assertRaises(TypeError):
            model.score_pair(sample_time_difference=np.array([2.5]), genetic_distance=1.5)

    def test_score_pair_requires_both_observations(self) -> None:
        profile = CountingProfile()
        model = EpiLink(
            transmission_profile=profile,
            maximum_depth=0,
            mc_samples=4,
            mutation_process="deterministic",
        )

        with self.assertRaises(TypeError):
            model.score_pair(sample_time_difference=2.5)

        with self.assertRaises(TypeError):
            model.score_pair(genetic_distance=1.5)

    def test_unknown_target_raises_value_error(self) -> None:
        profile = CountingProfile()

        with self.assertRaises(ValueError):
            EpiLink(
                transmission_profile=profile,
                maximum_depth=0,
                mc_samples=4,
                target="ad(3)",
            )

    def test_empty_target_subset_raises_value_error(self) -> None:
        profile = CountingProfile()

        with self.assertRaises(ValueError):
            EpiLink(
                transmission_profile=profile,
                maximum_depth=0,
                mc_samples=4,
                target=[],
            )


if __name__ == "__main__":
    unittest.main()
